/*
  Copyright 2008 CNRS
  Author: Florent Lamiraux
*/

#include <iostream>
#include "KineoWorks2/kwsDirectPath.h"
#include "KineoWorks2/kwsConfig.h"

#include "flicSteeringMethod.h"
#include "kwsPlusDirectPath.h"

#include "hppWalkFootPlanner/hppWalkDistance.h"

// Select level of verbosity at configuration.
#if DEBUG==2
#define ODEBUG2(x) std::cout << "ChppWalkDistance:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkDistance:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "ChppWalkDistance:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

ChppWalkDistance::~ChppWalkDistance()
{
}

ChppWalkDistanceShPtr ChppWalkDistance::create(double inDistanceThreshold)
{
  ChppWalkDistance* distance = new ChppWalkDistance(inDistanceThreshold);
  ChppWalkDistanceShPtr distanceShPtr(distance);
  ChppWalkDistanceWkPtr distanceWkPtr(distanceShPtr);

  if (distance->init(distanceWkPtr) == KD_ERROR) {
    distanceShPtr.reset();
  }
  return distanceShPtr;
}

double ChppWalkDistance::distance(const CkwsConfig &inConfig1, 
				  const CkwsConfig &inConfig2) const
{
  double distance = attFlicDistance->distance(inConfig1, inConfig2);

  // distance == FLIC_HUGE_VAL means that flic direct path between inConfig1 and inConfig2
  // does not exist.
  if (distance < FLIC_HUGE_VAL) {
    return distance;
  }

  // If flic direct path does not exist, return linear distance
  distance = attLinearDistance->distance(inConfig1, inConfig2);
  if (distance <= attDistanceThreshold) {
    return distance;
  }
  return FLIC_HUGE_VAL;
}

ktStatus ChppWalkDistance::init(const ChppWalkDistanceWkPtr &inWeakPtr)
{
  if (!attFlicDistance || !attLinearDistance) {
    return KD_ERROR;
  }

  ktStatus success = CkwsDistance::init(inWeakPtr);

  if (KD_OK == success) {
    attWeakPtr = inWeakPtr;
  }

  return success;

}

ChppWalkDistance::ChppWalkDistance(double inDistanceThreshold) :
  attWeakPtr(), 
  attLinearDistance(CkwsDistance::create()),
  attDistanceThreshold(inDistanceThreshold)
{
  CflicSteeringMethodShPtr sm = CflicSteeringMethod::create(true);
  attFlicDistance = CflicDistance::create(sm); 
}

/*
  The cost computed by this function penalizes motions based on the orientation 
  of the velocity vector of the center of the bounding box with respect to 
  the orientation of the box.
*/
double ChppWalkDistance::cost(const CkwsDirectPathConstShPtr &inDirectPath) const
{
  CkwsPlusDirectPathConstShPtr path = KIT_DYNAMIC_PTR_CAST(const CkwsPlusDirectPath, inDirectPath);
  double length = inDirectPath->length();
  
  static unsigned int halfNbSamples=20;
  static unsigned int nbSamples=2*halfNbSamples;
  double integrand[50];
  
  for (unsigned int iSample=0; iSample<nbSamples+1; iSample++) {
    double u=length*iSample/nbSamples;
    std::vector<double> velocity(inDirectPath->device()->countDofs());
    CkwsConfig config(inDirectPath->device());
    
    if (path->getVelocityAtDistance(u, velocity) == KD_ERROR) {
      ODEBUG1("cost: failed to get velocity at distance " << u);
      ODEBUG1("cost: length = " << length);
      return FLIC_HUGE_VAL;
    }
    if (path->getConfigAtDistance(u, config) ==KD_ERROR) {
      ODEBUG1("cost: failed to get config at distance " << u);
      ODEBUG1("cost: length = " << length);
      return FLIC_HUGE_VAL;
    }
    
    double vx = velocity[1];
    double vy = velocity[2];
    double theta = config.dofValue(6);
    
    integrand[iSample] = 2*(sqrt(vx*vx+vy*vy) - (vx*cos(theta) + vy*sin(theta)));
  }
  /*
    Simpson Formula
  */
  double sum = integrand[nbSamples] - integrand[0];
  //  ODEBUG2("iSample = 0, v = " << integrand[0]);
  //  ODEBUG2("iSample = " << nbSamples << ", v = " << integrand[nbSamples]);
  for (unsigned int iSample=0; iSample<halfNbSamples; iSample++) {
    sum += 2*integrand[2*iSample] + 4*integrand[2*iSample+1];
    //    ODEBUG2("iSample = " << 2*iSample << ", v = " << integrand[2*iSample]);
    //    ODEBUG2("iSample = " << 2*iSample+1 << ", v = " << integrand[2*iSample+1]);
  }
  double cost = 6*sum/(3*nbSamples) + length;
  
  return cost;
}

